import copy

from .pkg_resources.lipds.LiPD_Library import *
from .pkg_resources.timeseries.Convert import *
from .pkg_resources.timeseries.TimeSeries_Library import *
from .pkg_resources.doi.doi_main import doi_main
from .pkg_resources.excel.excel_main import excel_main
from .pkg_resources.noaa.noaa_main import noaa_prompt, noaa_to_lpd, lpd_to_noaa
from .pkg_resources.helpers.ts import translate_expression, get_matches
from .pkg_resources.helpers.dataframes import *
from .pkg_resources.helpers.directory import get_src_or_dst, collect_metadata_files, list_files
from .pkg_resources.helpers.loggers import create_logger
from .pkg_resources.helpers.misc import path_type, load_fn_matches_ext
from .pkg_resources.helpers.ensembles import create_ensemble, insert_ensemble
from .pkg_resources.helpers.validator_api import get_validator_results, display_results

# READ


def run():
    """
    Initialize and start objects
    :return:
    """
    # GLOBALS
    global cwd, lipd_lib, ts_lib, convert, files, logger_start, verbose
    verbose = True
    cwd = os.getcwd()
    # creating the LiPD Library also creates the Tmp sys folder where the files will be stored later.
    lipd_lib = LiPD_Library()
    ts_lib = TimeSeries_Library()
    convert = Convert()
    # print(cwd)
    # logger created in whatever directory lipd is called from
    logger_start = create_logger("start")
    files = {".txt": [], ".lpd": [], ".xls": []}
    return


def readLipd(usr_path=""):
    """
    Wrapper function for LiPD file read.
    :param str usr_path: Path to file
    :return none:
    """
    global cwd
    __read_file(usr_path, ".lpd")
    return cwd


def readLipds(usr_path=""):
    """
    Wrapper function for LiPD directory read.
    :param str usr_path: Path to directory
    :return none:
    """
    global cwd
    __read_directory(usr_path, ".lpd", "LiPD")
    # Turned off until LiPD.net is updated to accept Validator API requests
    # validator()
    return cwd


def readExcel(usr_path=""):
    """
    Wrapper function for Excel file read.
    :param str usr_path: Path to file
    :return none:
    """
    global cwd
    __read_file(usr_path, ".xls")
    return cwd


def readExcels(usr_path=""):
    """
    Wrapper function for Excel directory read.
    :param str usr_path: Path to directory
    :return none:
    """
    global cwd
    if not usr_path:
        usr_path, src_files = get_src_or_dst("read", "directory")
    __read_directory(usr_path, ".xls", "Excel (.xls)")
    __read_directory(usr_path, ".xlsx", "Excel (.xlsx)")
    return cwd


def readNoaa(usr_path=""):
    """
    Wrapper function for NOAA file read.
    :param str usr_path: Path to file
    :return none:
    """
    global cwd
    __read_file(usr_path, ".txt")
    return cwd


def readNoaas(usr_path=""):
    """
    Wrapper function for NOAA directory read.
    :param str usr_path: Path to directory
    :return none:
    """
    global cwd
    __read_directory(usr_path, ".txt", "NOAA")
    return cwd


def readAll(usr_path=""):
    """
    Wrapper function for reading ALL file types in a given directory.
    :param str usr_path: Path to directory
    :return none:
    """
    global cwd
    if not usr_path:
        usr_path, src_files = get_src_or_dst("read", "directory")
    __read_directory(usr_path, ".lpd", "LiPD")
    __read_directory(usr_path, ".xls", "Excel (xls)")
    __read_directory(usr_path, ".xlsx", "Excel (xlsx)")
    __read_directory(usr_path, ".txt", "NOAA")
    return cwd


def excel():
    """
    User facing call to the excel function
    :return none:
    """
    global files, cwd, verbose
    excel_main(files)
    # Turn off verbose. We don't want to clutter the console with extra reading/writing output statements
    verbose = False
    for file in files[".xls"]:
        try:
            readLipd(os.path.join(file["dir"], file["filename_no_ext"] + ".lpd"))
        except Exception as e:
            logger_start.debug("excel: Converted excel to lipd file, but unable readLipd(): {}, {}".format(file["filename_ext"], e))
    try:
        writeLipds(cwd)
    except Exception as e:
        logger_start.debug("excel: Unable to writeLipds with new lipd files, {}".format(e))
    # Turn on verbose. Back to normal mode when this function is finished.
    verbose = True
    return


def noaa():
    """
    User facing call to noaa function
    :return none:
    """
    global files, lipd_lib, cwd
    # When going from NOAA to LPD, use the global "files" variable.
    # When going from LPD to NOAA, use the data from the LiPD Library.

    # Choose the mode
    _mode = noaa_prompt()

    # LiPD mode: Convert LiPD files to NOAA files
    if _mode == "1":

        _lib = lipd_lib.get_master()

        # For each LiPD file in the LiPD Library
        for filename, obj in _lib.items():
            # Get the LiPD object data
            # _obj = dict(obj)
            # Process this data through the converter
            _obj_modified = lpd_to_noaa(obj)
            # Overwrite the data in the LiPD object with our new data.
            _lib[filename] = _obj_modified
        # Replace the data in the LiPD Library master
        lipd_lib.put_master(_lib)
        # Write out the new LiPD files, since they now contain the NOAA URL data
        writeLipds(cwd)

    # NOAA mode: Convert NOAA files to LiPD files
    elif _mode == "2":
        # For each NOAA txt file in the global "files" list, Run the converter
        #todo This is just a concept of how it should work. Not currently working.
        for file in files[".txt"]:
            _data = noaa_to_lpd(files)
            #todo Use the parsed NOAA data to create a LiPD object in the library
            #todo Use writeLipd to create a LiPD file using the object.
            pass

    else:
        print("Invalid input. Try again.")

    # LIPD TO NOAA IS THE ONLY CONVERSION CURRENTLY WORKING.
    # MAKE IT THE ONLY OPTION FOR NOW.
    # For each LiPD file in the LiPD Library
        # Get the LiPD object data
        # Process this data through the converter
        # Overwrite the data in the LiPD object with our new data.
        # Write the new LiPD file back out to disk.
    return


def doi():
    """
    User facing call to doi function
    :return:
    """
    global files
    doi_main(files)
    return


def validator(detailed=False):
    """
    Use the Validator API for lipd.net to validate all LiPD files in the LiPD Library.
    Display the PASS/FAIL results. Display detailed results if the option is chosen.
    :param bool detailed: Show or hide the detailed results of each LiPD file. Shows warnings and errors.
    :return none:
    """
    # Get the validator-formatted data for each LiPD file in the LiPD Library. A list of lists of LiPD-content metadata
    d = lipd_lib.get_data_for_validator()
    results = get_validator_results(d)
    display_results(results, detailed)
    return


# PUT


def addEnsemble(filename, ensemble):
    """
    Create ensemble entry and then add it to the specified LiPD dataset
    :param str filename: LiPD dataset name to add ensemble to
    :param list ensemble: Nested numpy array of ensemble column data.
    :return none:
    """
    # Get the master lipd library
    lib = lipd_lib.get_master()
    # Check that the given filename exists in the library
    if filename in lib:
        meta = lib[filename].get_metadata()
        # Create an ensemble dictionary entry
        ens = create_ensemble(ensemble)
        # If everything above worked, then there should be formatted ensemble data now.
        if ens:
            # Insert the formatted ensemble data into the master lipd library
            meta = insert_ensemble(meta, ens)
            # Set meta into lipd object
            lib[filename].put_metadata(meta)
            # Set the new master data back into the lipd library
            lipd_lib.put_master(lib)
    else:
        print("Error: '{}' file not found in workspace. Please use showLipds() to see available files ".format(filename))

    print("Process Complete")
    return


# DATA FRAMES


def ensToDf(ensemble):
    """
    Create an ensemble data frame from some given nested numpy arrays
    :param list ensemble: Ensemble data
    :return obj: DataFrame
    """
    df = create_dataframe(ensemble)
    return df


def lipdToDf(filename):
    """
    Get LiPD data frames from LiPD object
    :param str filename:
    :return dict: Pandas data frame objects
    """
    try:
        dfs = lipd_lib.get_dfs(filename)
    except KeyError:
        print("Error: Unable to find LiPD file")
        logger_start.warn("lipd_to_df: KeyError: missing lipds {}".format(filename))
        dfs = None
    print("Process Complete")
    return dfs


# todo this function no longer works. time series objects are not referenced by name. TS is a list of TSOs
def tsToDf(ts, filename):
    """
    Create Pandas DataFrame from TimeSeries object.
    Use: Must first extractTimeSeries to get a time series. Then pick one item from time series and pass it through
    :param dict ts: TimeSeries
    :param str filename:
    :return dict: Pandas data frames
    """
    dfs = {}
    try:
        dfs = ts_to_df(ts[filename])
    except KeyError as e:
        print("Error: LiPD file not found")
        logger_start.warn("ts_to_df: KeyError: LiPD file not found: {}".format(filename, e))
    print("Process Complete")
    return dfs


def filterDfs(expr):
    """
    Get data frames based on some criteria. i.e. all measurement tables or all ensembles.
    :param str expr: Search expression. (i.e. "paleo measurement tables")
    :return dict: Data frames indexed by filename
    """
    try:
        dfs = get_filtered_dfs(lipd_lib.get_master(), expr)
        print("Process Complete")
        return dfs

    except Exception:
        logger_dataframes.info("filter_dfs: Unable to filter data frames for expr: {}".format(expr))
        print("Unable to filter data frames")
        print("Process Complete")


# ANALYSIS - TIME SERIES


def extractTs():
    """
    Create a TimeSeries using the current files in LiPD_Library.
    :return list: TimeSeries_Library
    """
    l = []
    try:
        # Loop over the LiPD files in the LiPD_Library
        for k, v in lipd_lib.get_master().items():
            # Get metadata from this LiPD object. Convert.
            # Receive a time series (list of time series objects) and add it to what we currently have.
            # Continue building time series until all datasets are processed.
            print("extracting: {}".format(k))
            l += (convert.ts_extract_main(v.get_master(), v.get_dfs_chron()))
        print("Finished time series: {} objects".format(len(l)))
    except KeyError as e:
        print("Error: Unable to extractTimeSeries")
        logger_start.debug("extractTimeSeries() failed at {}".format(e))

    print("Process Complete")
    return l


def collapseTs():
    """
    Export TimeSeries back to LiPD Library. Updates information in LiPD objects.
    """
    l = []
    # Get all TSOs from TS_Library, and add them to a list
    for k, v in ts_lib.get_master().items():
        l.append({'name': v.get_lpd_name(), 'data': v.get_master()})
    # Send the TSOs list through to be converted. Then let the LiPD_Library load the metadata into itself.
    try:
        lipd_lib.load_tsos(convert.lipd_extract_main(l))
    except Exception:
        print("ERROR: Converting TSOs to LiPD")
        logger_start.debug("exportTimeSeries() failed")
    print("Process Complete")
    return


def find(expression, ts):
    """
    Find the names of the TimeSeries that match some criteria (expression)
    :return:
    """
    new_ts = []
    # filtered_ts = {}
    expr_lst = translate_expression(expression)
    if expr_lst:
        new_ts = get_matches(expr_lst, ts)
        # filtered_ts = _createTs(names, ts)
    print("Process Complete")
    return new_ts


# DEPRECATED - TS no longer uses dictionaries or names.
# def _createTs(names, ts):
#     """
#     Create a new TS dictionary using
#     index = find(logical expression)
#     newTS = TS(index)
#     :param str expression:
#     :return dict:
#     """
#     d = {}
#     for name in names:
#         try:
#             d[name] = ts[name]
#         except KeyError as e:
#             logger_start.warn("TS: KeyError: {} not in timeseries, {}".format(name, e))
#     return d


# SHOW


def showLipds():
    """
    Prints the names of all LiPD files in the LiPD_Library
    :return None:
    """
    lipd_lib.show_lipds()
    print("Process Complete")
    return


def showMetadata(filename):
    """
    Display the contents of the specified LiPD file. (Must be previously loaded into the workspace)
    (ex. displayLiPD NAm-ak000.lpd)
    :param str filename: LiPD filename
    """
    lipd_lib.show_metadata(filename)
    print("Process Complete")
    return


def showCsv(filename):
    """
    Show CSV data for one LiPD
    :param str filename:
    :return None:
    """
    lipd_lib.show_csv(filename)
    print("Process Complete")
    return


def showTso(tso):
    """
    Show contents of one TimeSeries object.
    :param str name: TimeSeries Object name
    :return None:
    """
    try:
        for key, value in sorted(tso.items()):
            print("{0:20}: {1}".format(key, value))
    except Exception as e:
        print("Unable to list contents: {}".format(e))
    print("Process Complete")
    return


def showDfs(dict_in):
    """
    Print the available data frame names in a given data frame collection
    :param dict dict_in: Data frame collection
    :return none:
    """
    if "metadata" in dict_in:
        print("metadata")
    if "paleoData" in dict_in:
        try:
            for k, v in dict_in["paleoData"].items():
                print(k)
        except KeyError:
            pass
        except AttributeError:
            pass
    if "chronData" in dict_in:
        try:
            for k, v in dict_in["chronData"].items():
                print(k)
        except KeyError:
            pass
        except AttributeError:
            pass
    print("Process Complete")
    return


# GET

def getLipdNames():
    """
    Get a list of all lipd dataset names in the library
    :return list:
    """
    f_list = []
    f_list = lipd_lib.get_lipd_names()
    return f_list


def getMetadata(filename):
    """
    Get metadata from LiPD file
    :param str filename: LiPD filename
    :return dict d: Metadata dictionary
    """
    d = {}
    try:
        d = lipd_lib.get_metadata(filename)
    except KeyError:
        print("Error: Unable to find LiPD file")
        logger_start.warn("KeyError: Unable to find record {}".format(filename))
    print("Process Complete")
    return d


def getCsv(filename):
    """
    Get CSV from LiPD file
    :param str filename: LiPD filename
    :return dict d: CSV dictionary
    """
    d = {}
    try:
        d = lipd_lib.get_csv(filename)
    except KeyError:
        print("Error: Unable to find record")
        logger_start.warn("Unable to find record {}".format(filename))
    print("Process Complete")
    return d


# WRITE


def writeLipd(filename, usr_path=""):
    """
    Write out one specific LiPD file
    :param str filename: LiPD filename ("somefile.lpd")
    :param str usr_path: Target directory destination (optional)
    :return none:
    """
    global verbose
    # no path provided. start gui browse
    if not usr_path:
        # got dir path
        usr_path, _ignore = get_src_or_dst("write", "directory")
    __write_lipd(usr_path, filename)
    return


def writeLipds(usr_path=""):
    """
    Save changes made to all LiPD files in the workspace.
    Files are created in current working directory if path is not provided.
    :param str usr_path: Target directory destination (optional)
    """
    global verbose
    # no path provided. start gui browse
    if not usr_path:
        # got dir path
        usr_path, _ignore = get_src_or_dst("write", "directory")
    _lib = list(lipd_lib.get_master().keys())
    if _lib:
        for filename in _lib:
            __write_lipd(usr_path, filename)
    if verbose:
        print("Process Complete")
    return


def removeLipd(filename):
    """
    Remove LiPD object from library
    :return None:
    """
    lipd_lib.remove_lipd(filename)
    return


def removeLipds():
    """
    Remove all LiPD objects from library.
    :return None:
    """
    lipd_lib.remove_lipds()
    return


def quit():
    """
    Quit and exit the program. (Does not save changes)
    """
    lipd_lib.cleanup()
    print("Quitting LiPD Utilities...")
    exit(0)


# HELPERS


def __universal_load(file_path, file_type):
    """
    Use a file path to create file metadata and load a file in the appropriate way, according to the provided file type.
    :param str file_path: Path to file
    :param str file_type: One of approved file types: xls, xlsx, txt, lpd
    :return none:
    """
    global files, lipd_lib, cwd, verbose

    # check that we are using the correct function to load this file type. (i.e. readNoaa for a .txt file)
    correct_ext = load_fn_matches_ext(file_path, file_type)

    # Check that this path references a file
    valid_path = path_type(file_path, "file")

    # is the path a file?
    if valid_path and correct_ext:

        # get file metadata for one file
        file_meta = collect_metadata_file(file_path)
        if verbose:
            print("reading: {}".format(file_meta["filename_ext"]))

        # append to global files, then load in lipd_lib
        if file_type == ".lpd":
            # add meta to global file meta
            files[".lpd"].append(file_meta)
            # yes, go ahead and load in the file
            lipd_lib.read_lipd(file_meta)
        # append to global files
        elif file_type in (".xls", ".xlsx"):
            files[".xls"].append(file_meta)
        # append to global files
        elif file_type == ".txt":
            files[".txt"].append(file_meta)

        # we want to move around with the files we load
        # change dir into the dir of the target file
        cwd = file_meta["dir"]
        os.chdir(cwd)

    return


def __read_file(usr_path, file_type):
    """
    Universal read file. Given a path and a type, it will do the appropriate read actions
    :param str usr_path: Path to file
    :param str file_type: One of approved file types: xls, xlsx, txt, lpd
    :return none:
    """
    global files

    # no path provided. start gui browse
    if not usr_path:
        # src files could be a list of one, or a list of many. depending how many files the user selects
        src_dir, src_files = get_src_or_dst("read", "file")
        # check if src_files is a list of multiple files
        if len(src_files) > 1:
            for file_path in src_files:
                __universal_load(file_path, file_type)
        # one file chosen
        elif src_files:
            file_path = src_files[0]
            __universal_load(file_path, file_type)
        else:
            print("No file(s) chosen")
    else:
        __universal_load(usr_path, file_type)

    return


def __read_directory(usr_path, file_type, file_type_print):
    """
    Universal read directory. Given a path and a type, it will do the appropriate read actions
    :param str usr_path: Path to directory
    :param str file_type: One of approved file types: xls, xlsx, txt, lpd
    :return none:
    """
    # no path provided. start gui browse
    if not usr_path:
        # got dir path
        usr_path, src_files = get_src_or_dst("read", "directory")

    # Check if this is a valid directory path
    valid_path = path_type(usr_path, "directory")

    # If dir path is valid
    if valid_path:
        # List all files of target type in dir
        files_found = list_files(file_type, usr_path)
        # notify how many files were found
        print("Found: {} {} file(s)".format(len(files_found), file_type_print))
        # Loop for each file found
        for file_path in files_found:
            # Call read lipd for each file found
            __read_file(file_path, file_type)
    else:
        print("Directory path is not valid: {}".format(usr_path))
    return


def __write_lipd(usr_path, filename):
    """
    Write LiPD data to file, provided an output directory and filename.
    :param str usr_path: Directory destination
    :param str filename: Target file
    :return none:
    """
    global verbose
    # no path provided. start gui browse
    if not usr_path:
        # got dir path
        usr_path, _ignore = get_src_or_dst("write", "directory")
    # Check if this is a valid directory path
    valid_path = path_type(usr_path, "directory")
    # If dir path is valid
    if valid_path:
        if verbose:
            print("writing: {}".format(filename))
        lipd_lib.write_lipd(usr_path, filename)

    return


# GLOBALS
run()




import operator
import os
import datetime as dt
import numpy as np
import unicodedata
import re
import shutil
import random
import string
import math

from ..helpers.loggers import create_logger
from ..helpers.blanks import EMPTY
from ..helpers.alternates import FILE_TYPE_MAP
from ..helpers.directory import list_files

logger_misc = create_logger("misc")


def cast_values_csvs(d, idx, x):
    """
    Attempt to cast string to float. If error, keep as a string.
    :param str x: String data
    :return any:
    """
    try:
        d[idx].append(float(x))
    except ValueError as e:
        d[idx].append(x)
        # logger_misc.warn("cast_values_csv: ValueError")
        # logger_misc.warn("ValueError: col: {}, {}".format(x, e))
    except KeyError as e:
        logger_misc.warn("cast_values_csv: KeyError: col: {}, {}".format(x, e))

    return d


def cast_float(x):
    """
    Attempt to cleanup string or convert to number value.
    :param any x:
    :return float:
    """
    try:
        x = float(x)
    except ValueError:
        try:
            x = x.strip()
        except AttributeError as e:
            logger_misc.warn("parse_str: AttributeError: String not number or word, {}, {}".format(x, e))
    return x


def cast_int(x):
    """
    Cast unknown type into integer
    :param any x:
    :return int:
    """
    try:
        x = int(x)
    except ValueError:
        try:
            x = x.strip()
        except AttributeError as e:
            logger_misc.warn("parse_str: AttributeError: String not number or word, {}, {}".format(x, e))
    return x


def clean_doi(doi_string):
    """
    Use regex to extract all DOI ids from string (i.e. 10.1029/2005pa001215)
    :param str doi_string: Raw DOI string value from input file. Often not properly formatted.
    :return list: DOI ids. May contain 0, 1, or multiple ids.
    """
    regex = re.compile(r'\b(10[.][0-9]{3,}(?:[.][0-9]+)*/(?:(?!["&\'<>,])\S)+)\b')
    try:
        # Returns a list of matching strings
        m = re.findall(regex, doi_string)
    except TypeError as e:
        # If doi_string is None type, return empty list
        logger_misc.warn("TypeError cleaning DOI: {}, {}".format(doi_string, e))
        m = []
    return m


def generate_timestamp(fmt=None):
    """
    Generate a timestamp to mark when this file was last modified.
    :param str fmt: Special format instructions
    :return str: YYYY-MM-DD format, or specified format
    """
    if fmt:
        time = dt.datetime.now().strftime(fmt)
    else:
        time = dt.date.today()
    return str(time)


def generate_tsid(size=8):
    """
    Generate a TSid string. Use the "PYT" prefix for traceability, and 8 trailing generated characters
    ex: PYT9AG234GS
    :return:
    """
    chars = string.ascii_uppercase + string.digits
    _gen = "".join(random.choice(chars) for _ in range(size))
    return "PYT" + str(_gen)


def get_authors_as_str(x):
    """
    Take author or investigator data, and convert it to a concatenated string of names.
    Author data structure has a few variations, so account for all.
    :param any x: Author data
    :return str: Author string
    """
    _authors = ""
    # if it's a string already, we're done
    if isinstance(x, str):
        return x

    # elif it's a list, keep going
    elif isinstance(x, list):
        # item in list is a str
        if isinstance(x[0], str):
            # loop and concat until the last item
            for name in x[:-1]:
                # all inner items get a semi-colon at the end
                _authors += str(name) + "; "
            # last item does not get a semi-colon at the end
            _authors += str(x[-1])

        # item in list is a dictionary
        elif isinstance(x[0], dict):
            # dictionary structure SHOULD have authors listed until the "name" key.
            try:
                # loop and concat until the last item
                for entry in x[:-1]:
                    # all inner items get a semi-colon at the end
                    _authors += str(entry["name"]) + "; "
                # last item does not get a semi-colon at the end
                _authors += str(x[-1]["name"])
            except KeyError:
                logger_misc.warn("get_authors_as_str: KeyError: Authors incorrect data structure")

    else:
        logger_misc.debug("get_authors_as_str: TypeError: author/investigators isn't str or list: {}".format(type(x)))

    return _authors


def get_missing_value_key(d, filename=""):
    """
    Get the Missing Value entry from a table of data. If none is found, try the columns. If still none found, prompt user.
    :param dict d: Table of data
    :param str filename: Filename linked to this table
    :return str: Missing Value
    """
    _mv = "nan"

    # Attempt to find a table-level missing value key
    try:
        # check for missing value key at the table root
        _mv = d["missingValue"]
    except KeyError as e:
        logger_misc.info("get_missing_value: No missing value key found: {}".format(e))
    except AttributeError as e:
        logger_misc.warn("get_missing_value: Column is wrong data type: {}".format(e))

    # No table-level missing value found. Attempt to find a column-level missing value key
    if not _mv:
        try:
            # loop for each column of data, searching for a missing value key
            for k, v in d["columns"].items():
                # found a column with a missing value key. Store it and exit the loop.
                _mv = v["missingValue"]
                break
        except KeyError:
            # There are no columns in this table. We've got bigger problems!
            pass

    # No table-level or column-level missing value. Out of places to look. Ask the user to enter the missing value
    # used in this data
    # if not _mv:
    #     print("No 'missingValue' key provided. Please type the missingValue used in this file: {}\n".format(filename))
    #     _mv = input("missingValue: ")
    return _mv


def get_variable_name_col(d):
    """
    Get the variable name from a table or column
    :param dict d: Metadata
    :return str:
    """
    var = ""
    try:
        var = d["variableName"]
    except KeyError:
        try:
            var = d["name"]
        except KeyError:
            num = "unknown"
            if "number" in d:
                num = d["number"]
            print("Error: column number <{}> is missing a variableName. Please fix.".format(num))
            logger_misc.info("get_variable_name_col: KeyError: missing key")
    return var


def get_variable_name_table(key, d, fallback=""):
    """
    Try to get a table name from a data table
    :param str key: Key to try first
    :param dict d: Data table
    :param str fallback: (optional) If we don't find a table name, use this as a generic name fallback.
    :return str: Data table name
    """
    try:
        var = d[key]
        return var
    except KeyError:
        logger_misc.info("get_variable_name_table: KeyError: missing {}".format(key))
        return fallback


def get_lipd_version(d):
    """
    Check what version of LiPD this file is using. If none is found, assume it's using version 1.0
    :param dict d: Metadata
    :return float:
    """
    version = 1.0
    if "LiPDVersion" in d:
        version = d["LiPDVersion"]
        # Cast the version number to a float
        try:
            version = float(version)
        except AttributeError:
            # If the casting failed, then something is wrong with the key so assume version is 1.0
            version = 1.0
    return version


def lipd_v1_0_to_v1_1(d):
    """
    Update LiPD version 1.0 to version 1.1.  See LiPD Version changelog for details.
    NOTE main changes: ChronData turned into a scalable lists of dictioanries.
    :param dict d: Metadata
    :return dict: v1.1 metadata dictionary
    """
    logger_misc.info("enter lipds 1.0 to 1.1")
    tmp_all = []

    # ChronData is the only structure update
    if "chronData" in d:
        # As of v1.1, ChronData should have an extra level of abstraction.
        # No longer shares the same structure of paleoData

        # If no measurement table, then make a measurement table list with the table as the entry
        for table in d["chronData"]:
            if "chronMeasurementTable" not in table:
                tmp_all.append({"chronMeasurementTable": [table]})

            # If the table exists, but it is a dictionary, then turn it into a list with one entry
            elif "chronMeasurementTable" in table:
                if isinstance(table["chronMeasurementTable"], dict):
                    tmp_all.append({"chronMeasurementTable": [table["chronMeasurementTable"]]})
        if tmp_all:
            d["chronData"] = tmp_all

    d["LiPDVersion"] = 1.1
    logger_misc.info("exit lipds 1.0 to 1.1")
    return d


def lipd_v1_1_to_v1_2(d):
    """
    Update LiPD version 1.1 to version 1.2. See LiPD Version changelog for details.
    :param dict d: Metadata dictioanry
    :return dict: v1.2 metadata dictionary
    """
    logger_misc.info("enter lipds 1.1 to 1.2")
    tmp_all = []

    # PaleoData is the only structure update
    if "paleoData" in d:

        # As of 1.2, PaleoData should match the structure of v1.1 chronData.
        # There is an extra level of abstraction and room for models, ensembles, calibrations, etc.
        for table in d["paleoData"]:
            if "paleoMeasurementTable" not in table:
                tmp_all.append({"paleoMeasurementTable": [table]})

            # If the table exists, but it is a dictionary, then turn it into a list with one entry
            elif "paleoMeasurementTable" in table:
                if isinstance(table["paleoMeasurementTable"], dict):
                    tmp_all.append({"paleoMeasurementTable": [table["paleoMeasurementTable"]]})

        if tmp_all:
            d["paleoData"] = tmp_all

    d["LiPDVersion"] = 1.2
    logger_misc.info("exit lipds 1.1 to 1.2")
    return d


def load_fn_matches_ext(file_path, file_type):
    """
    Check that the file extension matches the target extension given.
    :param str file_path: Path to be checked
    :param str file_type: Target extension
    :return bool:
    """
    correct_ext = False
    curr_ext = os.path.splitext(file_path)[1]
    exts = [curr_ext, file_type]
    try:
        # special case: if file type is excel, both extensions are valid.
        if ".xlsx" in exts and ".xls" in exts:
            correct_ext = True
        elif curr_ext == file_type:
            correct_ext = True
        else:
            print("Use '{}' to load this file: {}".format(FILE_TYPE_MAP[curr_ext]["load_fn"], os.path.basename(file_path)))
    except Exception as e:
        logger_misc.debug("load_fn_matches_ext: {}".format(e))

    return correct_ext


def match_operators(inp, relate, cut):
    """
    Compare two items. Match a string operator to an operator function
    :param str inp: Comparison item
    :param str relate: Comparison operator
    :param any cut: Comparison item
    :return bool: Comparison truth
    """
    logger_misc.info("enter match_operators")
    ops = {'>': operator.gt,
           '<': operator.lt,
           '>=': operator.ge,
           '<=': operator.le,
           '=': operator.eq
           }
    try:
        truth = ops[relate](inp, cut)
    except KeyError as e:
        truth = False
        logger_misc.warn("get_truth: KeyError: Invalid operator input: {}, {}".format(relate, e))
    logger_misc.info("exit match_operators")
    return truth


def match_arr_lengths(l):
    """
    Check that all the array lengths match so that a DataFrame can be created successfully.
    :param list l: Nested arrays
    :return bool: Valid or invalid
    """
    try:
        # length of first list. use as basis to check other list lengths against.
        inner_len = len(l[0])
        # check each nested list
        for i in l:
            # if the length doesn't match the first list, then don't proceed.
            if len(i) != inner_len:
                return False
    except IndexError:
        # couldn't get index 0. Wrong data type given or not nested lists
        print("Error: Array data is not formatted correctly.")
        return False
    except TypeError:
        # Non-iterable data type given.
        print("Error: Array data missing")
        return False
    # all array lengths are equal. made it through the whole list successfully
    return True


def mv_files(src, dst):
    """
    Move all files from one directory to another
    :param str src: Source directory
    :param str dst: Destination directory
    :return none:
    """
    # list the files in the src directory
    files = os.listdir(src)
    # loop for each file found
    for file in files:
        # move the file from the src to the dst
        shutil.move(os.path.join(src, file), os.path.join(dst, file))
    return


def normalize_name(s):
    """
    Remove foreign accents and characters to normalize the string. Prevents encoding errors.
    :param str s:
    :return str:
    """
    # Normalize the string into a byte string form
    s = unicodedata.normalize('NFKD', s).encode('ascii', 'ignore')
    # Remove the byte string and quotes from the string
    s = str(s)[2:-1]
    return s


def path_type(path, target):
    """
    Determine if given path is file, directory, or other. Compare with target to see if it's the type we wanted.
    :param str path: Path
    :param str target: Target type wanted
    :return bool:
    """
    if os.path.isfile(path) and target == "file":
        return True
    elif os.path.isdir(path) and target == "directory":
        return True
    else:
        print("Error: Path given is not a {}: {}".format(target, path))
    return False


def _prompt_filename():
    """
    Ask user if they want to provide a filename, or choose a generic time-stamped filename.
    :return:
    """
    filename = ""
    return filename


def prompt_protocol():
    """
    Prompt user if they would like to save pickle file as a dictionary or an object.
    :return str: Answer
    """
    stop = 3
    ans = ""
    while True and stop > 0:
        ans = input("Save as (d)ictionary or (o)bject?\n"
                      "* Note:\n"
                    "Dictionaries are more basic, and are compatible with Python v2.7+.\n"
                      "Objects are more complex, and are only compatible with v3.4+ ")
        if ans not in ("d", "o"):
            print("Invalid response: Please choose 'd' or 'o'")
        else:
            break
    # if a valid answer isn't captured, default to dictionary (safer, broader)
    if ans == "":
        ans = "d"
    return ans


def put_tsids(x):
    """
    Recursively add in TSids into any columns that do not have them.
    Look for "columns" keys, and then start looping and adding generated TSids to each column
    :param any x: Recursive, so could be any data type.
    :return any x: Recursive, so could be any data type.
    """
    try:
        if isinstance(x, dict):
            try:
                for k, v in x.items():
                    # Is this the columns key?
                    if k == "columns":
                        try:
                            # loop over each column of data. Sorted by variableName key
                            for var, data in v.items():
                                try:
                                    # make a case-insensitive keys list for checking existence of "tsid"
                                    keys = [key.lower() for key in data.keys()]
                                    # If a TSid already exists, then we don't need to do anything.
                                    if "tsid" not in keys:
                                        # generate the TSid, and add it to the dictionary
                                        data["TSid"] = generate_tsid()
                                        logger_misc.info("put_tsids: Generated new TSid: {}".format(data["TSid"]))
                                except AttributeError as e:
                                    logger_misc.debug("put_tsids: level 3: AttributeError: {}".format(e))
                                except Exception as e:
                                    logger_misc.debug("put_tsids: level 3: Exception: {}".format(e))
                        except Exception as e:
                            print("put_tsids: level 2: Exception: {}, {}".format(e, k))
                    # If it's not "columns", then dive deeper.
                    else:
                        x[k] = put_tsids(v)
            except Exception as e:
                print("put_tsids: level 1: Exception: {}, {}".format(e, k))
        # Item is a list, dive deeper for each item in the list
        elif isinstance(x, list):
            for idx, entry in enumerate(x):
                x[idx] = put_tsids(entry)
    except Exception as e:
        print("put_tsids: root: Exception: {}, {}".format(e, k))
    return x


def rm_empty_fields(x):
    """
    Go through N number of nested data types and remove all empty entries. Recursion
    :param any x: Dictionary, List, or String of data
    :return any: Returns a same data type as original, but without empties.
    """
    # No logger here because the function is recursive.
    # Int types don't matter. Return as-is.
    if not isinstance(x, int) and not isinstance(x, float):
        if isinstance(x, str) or x is None:
            try:
                # Remove new line characters and carriage returns
                x = x.rstrip()
            except AttributeError:
                # None types don't matter. Keep going.
                pass
            if x in EMPTY:
                # Substitute empty entries with ""
                x = ''
        elif isinstance(x, list):
            # Recurse once for each item in the list
            for i, v in enumerate(x):
                x[i] = rm_empty_fields(x[i])
            # After substitutions, remove empty entries.
            for i in x:
                # In the case of coordinates, we cannot remove coordinates of "0"
                if not i and i not in (0, 0.0):
                    x.remove(i)
        elif isinstance(x, dict):
            # First, go through and substitute "" (empty string) entry for any values in EMPTY
            for k, v in x.items():
                x[k] = rm_empty_fields(v)
            # After substitutions, go through and delete the key-value pair.
            # This has to be done after we come back up from recursion because we cannot pass keys down.
            for key in list(x.keys()):
                if not x[key]:
                    del x[key]
    return x


def rm_empty_doi(d):
    """
    If an "identifier" dictionary has no doi ID, then it has no use. Delete it.
    :param dict d: JSON Metadata
    :return dict: JSON Metadata
    """
    logger_misc.info("enter remove_empty_doi")
    try:
        # Check each publication dictionary
        for pub in d['pub']:
            # If no identifier, then we can quit here. If identifier, then keep going.
            if 'identifier' in pub:
                if 'id' in pub['identifier'][0]:
                    # If there's a DOI id, but it's EMPTY
                    if pub['identifier'][0]['id'] in EMPTY:
                        del pub['identifier']
                else:
                    # If there's an identifier section, with no DOI id
                    del pub['identifier']
    except KeyError as e:
        # What else could go wrong?
        logger_misc.warn("remove_empty_doi: KeyError: publication key not found, {}".format(e))
    logger_misc.info("exit remove_empty_doi")
    return d


def rm_files(path, extension):
    """
    Remove all files in the given directory with the given extension
    :param str path: Directory
    :param str extension: File type to remove
    :return none:
    """
    files = list_files(extension, path)
    for file in files:
        if file.endswith(extension):
            os.remove(os.path.join(path, file))
    return


def rm_values_fields(x):
    """
    (recursive) Remove all "values" fields from the metadata
    :param any x: Any data type
    :return dict: metadata without "values"
    """
    if isinstance(x, dict):
        if "values" in x:
            del x["values"]
        else:
            for k, v in x.items():
                if isinstance(v, dict):
                    rm_values_fields(v)
                elif isinstance(v, list):
                    rm_values_fields(v)
    elif isinstance(x, list):
        for i in x:
            rm_values_fields(i)

    return x


def rm_missing_values_table(d):
    """
    Loop for each table column and remove the missingValue key & data
    :param dict d: Table data
    :return dict d: Table data
    """
    try:
        for k, v in d["columns"].items():
            d["columns"][k] = rm_keys_from_dict(v, ["missingValue"])
    except Exception:
        # If we get a KeyError or some other error, it's not a big deal. Keep going.
        pass
    return d


def rm_keys_from_dict(d, keys):
    """
    Given a dictionary and a key list, remove any data in the dictionary with the given keys.
    :param dict d: Data
    :param list keys: List of key data to remove
    :return dict d: Data (with keys + data removed)
    """
    # Loop for each key given
    for key in keys:
        # Is the key in the dictionary?
        if key in d:
            try:
                d.pop(key, None)
            except KeyError:
                # Not concerned with an error. Keep going.
                pass
    return d


def _replace_missing_values_table(values, mv):
    """
    Receive all table column values as a list of lists. Loop for each column of values
    :param list values: One list per column
    :param any mv: Missing Value being used
    :return list: List of lists with updated "nan" missing values
    """

    for idx, column in enumerate(values):
        values[idx] = _replace_missing_values_column(column, mv)

    return values


def _replace_missing_values_column(values, mv):
    """
    Replace Missing Values in the values list where applicable
    :param list values: One column of values
    :return list: Column with updated "nan" missing values
    """
    for idx, v in enumerate(values):
        try:
            if v in EMPTY or v == mv:
                values[idx] = "nan"
            elif math.isnan(float(v)):
                values[idx] = "nan"
            else:
                values[idx] = v
        except (TypeError, ValueError):
            values[idx] = v

    return values


def split_path_and_file(s):
    """
    Given a full path to a file, split and return a path and filename
    :param str s: Full path
    :return str str: Path, filename
    """
    _path = s
    _filename = ""
    try:
        x = os.path.split(s)
        _path = x[0]
        _filename = x[1]
    except Exception:
        print("Error: unable to split path")

    return _path, _filename


def update_lipd_version(d):
    """
    Use the current version number to determine where to start updating from. Use "chain versioning" to make it
    modular. If a file is a few versions behind, convert to EACH version until reaching current. If a file is one
    version behind, it will only convert once to the newest.
    :param dict d: Metadata dictionary
    :return dict: Most current version metadata dictionary
    """

    # Get the lipds version number.
    version = get_lipd_version(d)

    # Update from (N/A or 1.0) to 1.1
    if version in (1.0, "1.0"):
        d = lipd_v1_0_to_v1_1(d)
        version = 1.1

    # Update from 1.1 to 1.2
    if version in (1.1, "1.1"):
        d = lipd_v1_1_to_v1_2(d)
        version = 1.2

    return d


def unwrap_arrays(l):
    """
    Unwrap nested lists to be one "flat" list of lists. Mainly for prepping ensemble data for DataFrame() creation
    :param list l: Nested lists
    :return list: Flattened lists
    """
    # keep processing until all nesting is removed
    process = True
    # fail safe: cap the loops at 20, so we don't run into an error and loop infinitely.
    # if it takes more than 20 loops then there is a problem with the data given.
    loops = 25
    while process and loops > 0:
        try:
            # new "flat" list
            l2 = []
            for k in l:
                # all items in this list are numeric, so this list is done. append to main list
                if all(isinstance(i, float) or isinstance(i, int) for i in k):
                        l2.append(k)
                # this list has more nested lists inside. append each individual nested list to the main one.
                elif all(isinstance(i, list) or isinstance(i, np.ndarray) for i in k):
                    for i in k:
                        l2.append(i)
        except Exception:
            print("something went wrong during process")
        # verify the main list
        try:
            # if every list has a numeric at index 0, then there is no more nesting and we can stop processing
            if all(isinstance(i[0], (int, str, float)) for i in l2):
                process = False
            else:
                l = l2
        except IndexError:
            # there's no index 0, so there must be mixed data types or empty data somewhere.
            print("something went wrong during verify")
        loops -= 1
    return l2




from ..helpers.loggers import *
from ..helpers.misc import unwrap_arrays, match_arr_lengths

logger_ensembles = create_logger("ensembles")


def create_ensemble(ensemble):
    """
    Add ensemble data to a LiPD object
    :param list ensemble: Ensemble data nested lists
    :return dict: Structured Ensemble data
    """
    logger_ensembles.info("enter add_ensemble")
    ens = {}
    # "Flatten" the nested lists. Bring all nested lists up to top-level. Output looks like [ [1,2], [1,2], ... ]
    ll = unwrap_arrays(ensemble)
    # Check that list lengths are all equal
    valid = match_arr_lengths(ll)
    if valid:
        # Template for ensemble structure
        ens = {"columns": {
                "depth":{"number": 1, "variableName": "depth", "units":"cm", "values": []},
                "age":{"number": [], "variableName": "age","units": "BP", "values": []}}}
        # Start creating the ensemble dictionary.
        for idx, l in enumerate(ll):
            if idx > 0:
                ens["columns"]["age"]["values"].append(l)
                ens["columns"]["age"]["number"].append(idx+1)
            else:
                ens["columns"]["depth"]["values"] = l
    else:
        # Lists are unequal. Print error and return nothing.
        print("Error: Numpy Array lengths do not match. Cannot create ensemble entry")
    logger_ensembles.info("exit add_ensemble")
    return ens


def insert_ensemble(d, ens):
    """
    Insert the ensemble table dictionary into the LiPD metadata
    :param dict d: LiPD metadata
    :param dict ens: Ensemble data to insert
    :return dict:
    """
    # Check for chronData section
    if "chronData" in d:
        # Have to loop, even though we'll quit after inserting on one table
        for k, v in d["chronData"].items():
            # Check for a chronModel
            if "chronModel" in v:
                # Check for a chronModel[0].
                try:
                    x = d["chronData"][k]["chronModel"][0]
                # No chronModel[0]. Build it.
                except IndexError:
                    d["chronData"][k]["chronModel"][0] = {}
                d["chronData"][k]["chronModel"][0]["ensembleTable"] = ens
            # No chronModel. Build it.
            else:
                v["chronModel"] = []
                v["chronModel"][0] = {}
                v["chronModel"][0]["ensembleTable"] = ens

    else:
        # There is no chronData section. Build from scratch.
        d["chronData"] = {}
        d["chronData"]["chron"] = {}
        d["chronData"]["chron"]["chronModel"] = []
        d["chronData"]["chron"]["chronModel"][0] = {"ensembleTable": ens}

    return d


# -*- coding: utf-8 -*-
import json
import requests
import os

from .loggers import create_logger

logger_validator_api = create_logger('validator_api')


def get_validator_format(data_json, data_csv, filenames):
    """
    Format the LIPD data in the layout that the Lipd.net validator accepts.
    Example of one _file metadata. _file_list will contain 1 or more _file's
    _file = {
        "type": "bagit/json/csv",
        "filenameFull": /path/to/filename.txt,
        "filenameShort": filename.txt,
        "data": "",
        "pretty": ""
    }

    :param dict data_json: Metadata
    :param dict data_csv: CSV data
    :param list filenames: All files found in LiPD archive
    :return list: Validator-formatted data
    """
    # Store the metadata for each file in the _file_list
    _file_lst = []
    # Loop for each filename
    for filename in filenames:
        # Create a blank template
        _file = {"type": "", "filenameFull": "", "filenameShort": "", "data": "", "pretty": ""}
        # filename, no path prefix
        _short = os.path.basename(filename)
        # Bagit files
        if filename.endswith(".txt"):
            _file = {"type": "bagit", "filenameFull": filename, "filenameShort": _short}
        # JSONLD files
        elif filename.endswith(".jsonld"):
            _file = {"type": "json", "filenameFull": filename, "filenameShort": _short, "data": data_json}

        # CSV files
        elif filename.endswith(".csv"):
            _cols = len(data_csv[os.path.basename(filename)])
            _rows = 0
            for k, v in data_csv[_short].items():
                _rows = len(v)
                break
            _file = {"type": "csv", "filenameFull": filename, "filenameShort": _short,
                     "data": {"cols": _cols, "rows": _rows}}
        _file_lst.append(_file)

    return _file_lst


def create_detailed_results(data):
    """
    """
    string = ""
    # Validation Response output
    # string += "VALIDATION RESPONSE\n"
    string += "STATUS: {}\n".format(data["status"])
    if data["feedback"]:
        string += "WARNINGS: {}\n".format(len(data["feedback"]["wrnMsgs"]))
        # Loop through and output the Warnings
        for msg in data["feedback"]["wrnMsgs"]:
            string += "- {}\n".format(msg)
        string += "ERRORS: {}\n".format(len(data["feedback"]["errMsgs"]))
        # Loop through and output the Errors
        for msg in data["feedback"]["errMsgs"]:
            string += "- {}\n".format(msg)
    return string


def display_results(data, detailed=False):
    """
    Display the results from the validator in a brief or detailed way.
    :param dict data: Results, sorted by dataset name
    :param bool detailed: Detailed results on or off
    :return none:
    """
    print("\nVALIDATOR RESULTS")
    print("======================\n")

    if not detailed:
        print('FILENAME......................................... STATUS..........')

    for lipd in data:
        try:
            if detailed:
                print("\n{}".format(lipd["filename"]))
                print(create_detailed_results(lipd))
            else:
                print("{:<50}{}".format(lipd["filename"], lipd["status"]))
        except Exception as e:
            logger_validator_api.debug("display_results: Exception: {}".format(e))
            print("skipping...")

    return


def get_validator_results(data):
    """
    Send LiPD data to the Lipd.net validator and get the results back.
    :param data:
    :return:
    """
    # results = {"detailed": "", "raw": {}, "status": "", "filename":""}
    results = []
    for file in data:
        # Add this single results to the growing list of results.
        result = _call_validator_api(file)
        results.append(result[0])

    return results


def _get_failed_filename(file):
    """
    File has failed JSON dump. Find the filename so we can notify the user.
    :param file:
    :return:
    """
    _filename = "unknown file"
    try:
        _filename = file[0]["filenameFull"].split("/")[0] + ".lpd"
    except IndexError:
        pass
    return _filename


def _call_validator_api(data):
    """
    Single call to the lipd.net validator API
    Returns data in format: {"dat": <dict>, "feedback": <dict>, "filename": "", "status": ""}

    :param list data: Prepared payload. JSON dumped data. Single or batch.
    :return list: Validator results
    """
    _filename = _get_failed_filename(data)

    try:
        # Contact server and send LiPD metadata as the payload
        # print("Sending request to LiPD.net validator...\n")
        data = json.dumps([data])

        # The payload that is going to be sent with the JSON request
        payload = {'json_payload': data, 'apikey': 'YOUR_API_KEY_HERE'}
        # Development Link

        response = requests.get('http://localhost:3000/api/validator', data=payload)
        # Production Link
        # response = requests.get('http://www.lipd.net/api/validate', data=payload)

        # For an example of the JSON Response, reference the "sample_data_response" below

        # Convert JSON string into a Python dictionary
        # print("Converting response to json...\n")
        result = json.loads(response.text)

    except TypeError as e:
        logger_validator_api.warning("get_validator_results: TypeError: {}".format(e))
        result = [{"dat": {}, "feedback": {}, "filename": _filename, "status": "DECODE ERROR"}]
    except requests.exceptions.ConnectionError as e:
        logger_validator_api.warning("get_validator_results: ConnectionError: {}".format(e))
        result = [{"dat": {}, "feedback": {}, "filename": _filename, "status": "UNABLE TO REACH SERVER"}]
    except Exception as e:
        logger_validator_api.debug("get_validator_results: Exception: {}".format(e))
        result = [{"dat": {}, "feedback": {}, "filename": _filename, "status": "ERROR BEFORE VALIDATION"}]
    if not result:
        result = [{"dat": {}, "feedback": {}, "filename": _filename, "status": "EMPTY RESPONSE"}]

    return result

# Validator Format
# This is a list of lists
# Each request contains one array
# That main array contains one array per LiPD archive
# Each LiPD archive array contains one JSON object per file inside of the LiPD archive (3 bagit txt, 1+ jsonld, 1+ csv)
sample_data_request = [

    [
        {
            "type": "bagit",
            "filenameFull": "MD98-2170.Stott.2004/bag-info.txt",
            "filenameShort": "bag-info.txt",
            "data": "",
            # "pretty": "\"\""
        },
        {
            "type": "bagit",
            "filenameFull": "MD98-2170.Stott.2004/bagit.txt",
            "filenameShort": "bagit.txt",
            "data": "BagIt-Version: 0.96\nTag-File-Character-Encoding: UTF-8\n",
            # "pretty": "\"BagIt-Version: 0.96\\nTag-File-Character-Encoding: UTF-8\\n\""
        },
        {
            "type": "bagit",
            "filenameFull": "MD98-2170.Stott.2004/manifest-md5.txt",
            "filenameShort": "manifest-md5.txt",
            "data": "36bc011ad51cb607b4c5387262f51320 data/MD98-2170.Stott.2004.chron1measurement1.csv\nee0a2f9444dee56be93d87c7c30114a2 data/MD98-2170.Stott.2004.jsonld\n5d6b9d190dac29878874a92a0f616169 data/MD98-2170.Stott.2004.paleo1measurement1.csv\n",
            # "pretty": "\"36bc011ad51cb607b4c5387262f51320 data/MD98-2170.Stott.2004.chron1measurement1.csv\\nee0a2f9444dee56be93d87c7c30114a2 data/MD98-2170.Stott.2004.jsonld\\n5d6b9d190dac29878874a92a0f616169 data/MD98-2170.Stott.2004.paleo1measurement1.csv\\n\""
        },
        {
            "type": "bagit",
            "filenameFull": "MD98-2170.Stott.2004/tagmanifest-md5.txt",
            "filenameShort": "tagmanifest-md5.txt",
            "data": "d41d8cd98f00b204e9800998ecf8427e bag-info.txt\nace0ef9419c8edbe164a888d4e4ab7ee bagit.txt\n6754ca6beb4812de50cec58d8b4b2ef9 manifest-md5.txt\n",
            # "pretty": "\"d41d8cd98f00b204e9800998ecf8427e bag-info.txt\\nace0ef9419c8edbe164a888d4e4ab7ee bagit.txt\\n6754ca6beb4812de50cec58d8b4b2ef9 manifest-md5.txt\\n\""
        },
        {
            "type": "csv",
            "filenameFull": "MD98-2170.Stott.2004/data/MD98-2170.Stott.2004.chron1measurement1.csv",
            "filenameShort": "MD98-2170.Stott.2004.chron1measurement1.csv",
            "data": {
                "data": [
                    [
                        "7.0",
                        "OS-38364",
                        "G. sacculifer",
                        "2370.0",
                        "25.0"
                    ],
                    [
                        "102.0",
                        "OS-38365",
                        "G. sacculifer",
                        "8520.0",
                        "40.0"
                    ],
                    [
                        "200.0",
                        "OS-38366",
                        "G. sacculifer",
                        "6900.0",
                        "35.0"
                    ],
                    [
                        "300.0",
                        "OS-38367",
                        "G. sacculifer",
                        "9710.0",
                        "45.0"
                    ],
                    [
                        "401.0",
                        "OS-54863",
                        "G. sacculifer",
                        "14900.0",
                        "90.0"
                    ],
                    [
                        "427.0",
                        "OS-54864",
                        "G. sacculifer",
                        "15450.0",
                        "80.0"
                    ],
                    [
                        ""
                    ]
                ],
                "errors": [],
                "meta": {
                    "delimiter": ",",
                    "linebreak": "\r\n",
                    "cursor": 252
                },
                "rows": 7,
                "cols": 5,
                "delimiter": ","
            },
            # "pretty": "{\n  \"data\": [\n    [\n      \"7.0\",\n      \"OS-38364\",\n      \"G. sacculifer\",\n      \"2370.0\",\n      \"25.0\"\n    ],\n    [\n      \"102.0\",\n      \"OS-38365\",\n      \"G. sacculifer\",\n      \"8520.0\",\n      \"40.0\"\n    ],\n    [\n      \"200.0\",\n      \"OS-38366\",\n      \"G. sacculifer\",\n      \"6900.0\",\n      \"35.0\"\n    ],\n    [\n      \"300.0\",\n      \"OS-38367\",\n      \"G. sacculifer\",\n      \"9710.0\",\n      \"45.0\"\n    ],\n    [\n      \"401.0\",\n      \"OS-54863\",\n      \"G. sacculifer\",\n      \"14900.0\",\n      \"90.0\"\n    ],\n    [\n      \"427.0\",\n      \"OS-54864\",\n      \"G. sacculifer\",\n      \"15450.0\",\n      \"80.0\"\n    ],\n    [\n      \"\"\n    ]\n  ],\n  \"errors\": [],\n  \"meta\": {\n    \"delimiter\": \",\",\n    \"linebreak\": \"\\r\\n\",\n    \"aborted\": false,\n    \"truncated\": false,\n    \"cursor\": 252\n  },\n  \"rows\": 7,\n  \"cols\": 5,\n  \"delimiter\": \",\"\n}"
        },
        {
            "type": "json",
            "filenameFull": "MD98-2170.Stott.2004/data/MD98-2170.Stott.2004.jsonld",
            "filenameShort": "MD98-2170.Stott.2004.jsonld",
            "data": {
                "archiveType": "marine sediment",
                "geo": {
                    "geometry": {
                        "coordinates": [
                            125.39,
                            -10.59,
                            -832
                        ],
                        "type": "Point"
                    },
                    "properties": {
                        "siteName": "MD98-2170"
                    },
                    "type": "Feature"
                },
                "investigator": [
                    {
                        "name": "L.D. Stott"
                    },
                    {
                        "name": "K.G. Cannariato"
                    },
                    {
                        "name": "R.C. Thunell"
                    },
                    {
                        "name": "G.H. Haug"
                    },
                    {
                        "name": "A. Koutavas"
                    },
                    {
                        "name": "S.P. Lund"
                    }
                ],
                "description": "Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions. delta O18 PDB (Globigerinoides ruber); Magnesium/Calcium ratio (Globigerinoides ruber); Sea Surface Temperature (C); delta O18 SMOW (water); radiocarbon years before 1950AD",
                "chronData": [
                    {
                        "chronMeasurementTable": [
                            {
                                "chronDataTableName": "chron1measurement1",
                                "filename": "MD98-2170.Stott.2004.chron1measurement1.csv",
                                "columns": [
                                    {
                                        "number": "1",
                                        "variableName": "depth",
                                        "variableType": "measured"
                                    },
                                    {
                                        "number": "2",
                                        "variableName": "labcode",
                                        "takenAtDepth": {
                                            "foundInTable": "MD98-2170.Stott.2004.chron1measurement1",
                                            "hasColumnNumber": "1",
                                            "name": "depth"
                                        },
                                        "variableType": "measured"
                                    },
                                    {
                                        "number": "3",
                                        "variableName": "mat",
                                        "takenAtDepth": {
                                            "foundInTable": "MD98-2170.Stott.2004.chron1measurement1",
                                            "hasColumnNumber": "1",
                                            "name": "depth"
                                        },
                                        "variableType": "measured"
                                    },
                                    {
                                        "number": "4",
                                        "units": "yr BP",
                                        "measuredOn": {
                                            "archiveName": "MD98-2170",
                                            "sensingWith": "Globigerinoides sacculifer"
                                        },
                                        "variableName": "age14c",
                                        "onProxyObservationProperty": "Radiocarbon",
                                        "takenAtDepth": {
                                            "foundInTable": "MD98-2170.Stott.2004.chron1measurement1",
                                            "hasColumnNumber": "1",
                                            "name": "depth"
                                        },
                                        "variableType": "measured",
                                        "uncertainty": 0,
                                        "hasUnits": "yr BP",
                                        "uncertaintyLevel": "1 sigma"
                                    },
                                    {
                                        "number": "5",
                                        "variableName": "age14cuncertainty",
                                        "takenAtDepth": ".depth",
                                        "variableType": "measured"
                                    }
                                ],
                                "missingValue": "NaN"
                            }
                        ]
                    }
                ],
                "paleoData": [
                    {
                        "paleoMeasurementTable": [
                            {
                                "filename": "MD98-2170.Stott.2004.paleo1measurement1.csv",
                                "columns": [
                                    {
                                        "description": "depth (cm)",
                                        "number": "1",
                                        "units": "cm",
                                        "variableName": "depth",
                                        "variableType": "measured"
                                    },
                                    {
                                        "description": "age yr BP",
                                        "number": "2",
                                        "missingValue": "-999",
                                        "units": "yr BP",
                                        "max": 14575.9,
                                        "min": 0,
                                        "variableName": "age",
                                        "onInferredVariableProperty": "Age",
                                        "takenAtDepth": {
                                            "description": "depth (cm)",
                                            "foundInTable": "MD98-2170.Stott.2004.paleo1measurement1",
                                            "hasColumnNumber": "1",
                                            "hasUnits": "cm",
                                            "name": "depth"
                                        },
                                        "variableType": "inferred"
                                    },
                                    {
                                        "description": "d18O PDB",
                                        "number": "3",
                                        "missingValue": "-999",
                                        "units": "per mil",
                                        "measuredOn": {
                                            "archiveName": "MD98-2170",
                                            "sensingWith": {
                                                "sensorGenus": "Globigerinoides",
                                                "sensorSpecies": "ruber"
                                            }
                                        },
                                        "measurementMaterial": "Globigerinoides ruber",
                                        "variableName": "d18o",
                                        "onProxyObservationProperty": "D18O",
                                        "measurementStandard": "VPDB",
                                        "takenAtDepth": {
                                            "description": "depth (cm)",
                                            "foundInTable": "MD98-2170.Stott.2004.paleo1measurement1",
                                            "hasColumnNumber": "1",
                                            "hasUnits": "cm",
                                            "name": "depth"
                                        },
                                        "variableType": "measured"
                                    },
                                    {
                                        "description": "d18O SMOW",
                                        "number": "4",
                                        "missingValue": "-999",
                                        "units": "per mil SMOW",
                                        "measurementMaterial": "water",
                                        "variableName": "d18ow",
                                        "onInferredVariableProperty": "D18Osw",
                                        "takenAtDepth": {
                                            "description": "depth (cm)",
                                            "foundInTable": "MD98-2170.Stott.2004.paleo1measurement1",
                                            "hasColumnNumber": "1",
                                            "hasUnits": "cm",
                                            "name": "depth"
                                        },
                                        "variableType": "inferred"
                                    },
                                    {
                                        "description": "Mg/Ca ratio",
                                        "number": "5",
                                        "missingValue": "-999",
                                        "units": "mmol/mol",
                                        "measuredOn": {
                                            "archiveName": "MD98-2170",
                                            "sensingWith": {
                                                "sensorGenus": "Globigerinoides",
                                                "sensorSpecies": "ruber"
                                            }
                                        },
                                        "variableName": "mg",
                                        "onProxyObservationProperty": "Mg/Ca",
                                        "takenAtDepth": {
                                            "description": "depth (cm)",
                                            "foundInTable": "MD98-2170.Stott.2004.paleo1measurement1",
                                            "hasColumnNumber": "1",
                                            "hasUnits": "cm",
                                            "name": "depth"
                                        },
                                        "variableType": "measured"
                                    },
                                    {
                                        "description": "Mg/Ca sea surface temperature",
                                        "number": "6",
                                        "missingValue": "-999",
                                        "units": "deg C",
                                        "variableName": "sst",
                                        "onInferredVariableProperty": "SST",
                                        "takenAtDepth": {
                                            "description": "depth (cm)",
                                            "foundInTable": "MD98-2170.Stott.2004.paleo1measurement1",
                                            "hasColumnNumber": "1",
                                            "hasUnits": "cm",
                                            "name": "depth"
                                        },
                                        "variableType": "inferred"
                                    }
                                ],
                                "missingValue": "-999",
                                "paleoDataTableName": "paleo1measurement1"
                            }
                        ]
                    }
                ],
                "liPDVersion": 1.2,
                "dataSetName": "MD98-2170.Stott.2004",
                "proxy": "Paleoceanography",
                "pub": [
                    {
                        "abstract": "In the present-day climate, surface water salinities are low in the western tropical Pacific Ocean and increase towards the eastern part of the basin. The salinity of surface waters in the tropical Pacific Ocean is thought to be controlled by a combination of atmospheric convection, precipitation, evaporation and ocean dynamics, and on interannual timescales significant variability is associated with the El Niņo/Southern Oscillation cycles. However, little is known about the variability of the coupled ocean-atmosphere system on timescales of centuries to millennia. Here we combine oxygen isotope and Mg/Ca data from foraminifers retrieved from three sediment cores in the western tropical Pacific Ocean to reconstruct Holocene sea surface temperatures and salinities in the region. We find a decrease in sea surface temperatures of ~0.5 °C over the past 10,000 yr, whereas sea surface salinities decreased by ~1.5 practical salinity units. Our data imply either that the Pacific basin as a whole has become progressively less salty or that the present salinity gradient along the Equator has developed relatively recently.",
                        "author": [
                            {
                                "name": "L.D. Stott"
                            },
                            {
                                "name": "K.G. Cannariato"
                            },
                            {
                                "name": "R.C. Thunell"
                            },
                            {
                                "name": "G.H. Haug"
                            },
                            {
                                "name": "A. Koutavas"
                            },
                            {
                                "name": "S.P. Lund"
                            }
                        ],
                        "journal": "Nature",
                        "pages": "56-59",
                        "pubDataUrl": "Manually Entered",
                        "title": "Decline of surface temperature and salinity in the western tropical Pacific Ocean in the Holocene epoch.",
                        "volume": "431",
                        "identifier": []
                    }
                ],
                "studyName": "Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions"
            },
            "pretty": "{\n  \"archiveType\": \"marine sediment\",\n  \"geo\": {\n    \"geometry\": {\n      \"coordinates\": [\n        125.39,\n        -10.59,\n        -832\n      ],\n      \"type\": \"Point\"\n    },\n    \"properties\": {\n      \"siteName\": \"MD98-2170\"\n    },\n    \"type\": \"Feature\"\n  },\n  \"investigator\": [\n    {\n      \"name\": \"L.D. Stott\"\n    },\n    {\n      \"name\": \"K.G. Cannariato\"\n    },\n    {\n      \"name\": \"R.C. Thunell\"\n    },\n    {\n      \"name\": \"G.H. Haug\"\n    },\n    {\n      \"name\": \"A. Koutavas\"\n    },\n    {\n      \"name\": \"S.P. Lund\"\n    }\n  ],\n  \"description\": \"Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions. delta O18 PDB (Globigerinoides ruber); Magnesium/Calcium ratio (Globigerinoides ruber); Sea Surface Temperature (C); delta O18 SMOW (water); radiocarbon years before 1950AD\",\n  \"chronData\": [\n    {\n      \"chronMeasurementTable\": [\n        {\n          \"chronDataTableName\": \"chron1measurement1\",\n          \"filename\": \"MD98-2170.Stott.2004.chron1measurement1.csv\",\n          \"columns\": [\n            {\n              \"number\": \"1\",\n              \"variableName\": \"depth\",\n              \"variableType\": \"measured\"\n            },\n            {\n              \"number\": \"2\",\n              \"variableName\": \"labcode\",\n              \"takenAtDepth\": {\n                \"foundInTable\": \"MD98-2170.Stott.2004.chron1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"measured\"\n            },\n            {\n              \"number\": \"3\",\n              \"variableName\": \"mat\",\n              \"takenAtDepth\": {\n                \"foundInTable\": \"MD98-2170.Stott.2004.chron1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"measured\"\n            },\n            {\n              \"number\": \"4\",\n              \"units\": \"yr BP\",\n              \"measuredOn\": {\n                \"archiveName\": \"MD98-2170\",\n                \"sensingWith\": \"Globigerinoides sacculifer\"\n              },\n              \"variableName\": \"age14c\",\n              \"onProxyObservationProperty\": \"Radiocarbon\",\n              \"takenAtDepth\": {\n                \"foundInTable\": \"MD98-2170.Stott.2004.chron1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"measured\",\n              \"uncertainty\": 0,\n              \"hasUnits\": \"yr BP\",\n              \"uncertaintyLevel\": \"1 sigma\"\n            },\n            {\n              \"number\": \"5\",\n              \"variableName\": \"age14cuncertainty\",\n              \"takenAtDepth\": \".depth\",\n              \"variableType\": \"measured\"\n            }\n          ],\n          \"missingValue\": \"NaN\"\n        }\n      ]\n    }\n  ],\n  \"paleoData\": [\n    {\n      \"paleoMeasurementTable\": [\n        {\n          \"filename\": \"MD98-2170.Stott.2004.paleo1measurement1.csv\",\n          \"columns\": [\n            {\n              \"description\": \"depth (cm)\",\n              \"number\": \"1\",\n              \"units\": \"cm\",\n              \"variableName\": \"depth\",\n              \"variableType\": \"measured\"\n            },\n            {\n              \"description\": \"age yr BP\",\n              \"number\": \"2\",\n              \"missingValue\": \"-999\",\n              \"units\": \"yr BP\",\n              \"max\": 14575.9,\n              \"min\": 0,\n              \"variableName\": \"age\",\n              \"onInferredVariableProperty\": \"Age\",\n              \"takenAtDepth\": {\n                \"description\": \"depth (cm)\",\n                \"foundInTable\": \"MD98-2170.Stott.2004.paleo1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"hasUnits\": \"cm\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"inferred\"\n            },\n            {\n              \"description\": \"d18O PDB\",\n              \"number\": \"3\",\n              \"missingValue\": \"-999\",\n              \"units\": \"per mil\",\n              \"measuredOn\": {\n                \"archiveName\": \"MD98-2170\",\n                \"sensingWith\": {\n                  \"sensorGenus\": \"Globigerinoides\",\n                  \"sensorSpecies\": \"ruber\"\n                }\n              },\n              \"measurementMaterial\": \"Globigerinoides ruber\",\n              \"variableName\": \"d18o\",\n              \"onProxyObservationProperty\": \"D18O\",\n              \"measurementStandard\": \"VPDB\",\n              \"takenAtDepth\": {\n                \"description\": \"depth (cm)\",\n                \"foundInTable\": \"MD98-2170.Stott.2004.paleo1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"hasUnits\": \"cm\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"measured\"\n            },\n            {\n              \"description\": \"d18O SMOW\",\n              \"number\": \"4\",\n              \"missingValue\": \"-999\",\n              \"units\": \"per mil SMOW\",\n              \"measurementMaterial\": \"water\",\n              \"variableName\": \"d18ow\",\n              \"onInferredVariableProperty\": \"D18Osw\",\n              \"takenAtDepth\": {\n                \"description\": \"depth (cm)\",\n                \"foundInTable\": \"MD98-2170.Stott.2004.paleo1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"hasUnits\": \"cm\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"inferred\"\n            },\n            {\n              \"description\": \"Mg/Ca ratio\",\n              \"number\": \"5\",\n              \"missingValue\": \"-999\",\n              \"units\": \"mmol/mol\",\n              \"measuredOn\": {\n                \"archiveName\": \"MD98-2170\",\n                \"sensingWith\": {\n                  \"sensorGenus\": \"Globigerinoides\",\n                  \"sensorSpecies\": \"ruber\"\n                }\n              },\n              \"variableName\": \"mg\",\n              \"onProxyObservationProperty\": \"Mg/Ca\",\n              \"takenAtDepth\": {\n                \"description\": \"depth (cm)\",\n                \"foundInTable\": \"MD98-2170.Stott.2004.paleo1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"hasUnits\": \"cm\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"measured\"\n            },\n            {\n              \"description\": \"Mg/Ca sea surface temperature\",\n              \"number\": \"6\",\n              \"missingValue\": \"-999\",\n              \"units\": \"deg C\",\n              \"variableName\": \"sst\",\n              \"onInferredVariableProperty\": \"SST\",\n              \"takenAtDepth\": {\n                \"description\": \"depth (cm)\",\n                \"foundInTable\": \"MD98-2170.Stott.2004.paleo1measurement1\",\n                \"hasColumnNumber\": \"1\",\n                \"hasUnits\": \"cm\",\n                \"name\": \"depth\"\n              },\n              \"variableType\": \"inferred\"\n            }\n          ],\n          \"missingValue\": \"-999\",\n          \"paleoDataTableName\": \"paleo1measurement1\"\n        }\n      ]\n    }\n  ],\n  \"liPDVersion\": 1.2,\n  \"dataSetName\": \"MD98-2170.Stott.2004\",\n  \"proxy\": \"Paleoceanography\",\n  \"pub\": [\n    {\n      \"abstract\": \"In the present-day climate, surface water salinities are low in the western tropical Pacific Ocean and increase towards the eastern part of the basin. The salinity of surface waters in the tropical Pacific Ocean is thought to be controlled by a combination of atmospheric convection, precipitation, evaporation and ocean dynamics, and on interannual timescales significant variability is associated with the El Niņo/Southern Oscillation cycles. However, little is known about the variability of the coupled ocean-atmosphere system on timescales of centuries to millennia. Here we combine oxygen isotope and Mg/Ca data from foraminifers retrieved from three sediment cores in the western tropical Pacific Ocean to reconstruct Holocene sea surface temperatures and salinities in the region. We find a decrease in sea surface temperatures of ~0.5 °C over the past 10,000 yr, whereas sea surface salinities decreased by ~1.5 practical salinity units. Our data imply either that the Pacific basin as a whole has become progressively less salty or that the present salinity gradient along the Equator has developed relatively recently.\",\n      \"author\": [\n        {\n          \"name\": \"L.D. Stott\"\n        },\n        {\n          \"name\": \"K.G. Cannariato\"\n        },\n        {\n          \"name\": \"R.C. Thunell\"\n        },\n        {\n          \"name\": \"G.H. Haug\"\n        },\n        {\n          \"name\": \"A. Koutavas\"\n        },\n        {\n          \"name\": \"S.P. Lund\"\n        }\n      ],\n      \"journal\": \"Nature\",\n      \"pages\": \"56-59\",\n      \"pubDataUrl\": \"Manually Entered\",\n      \"title\": \"Decline of surface temperature and salinity in the western tropical Pacific Ocean in the Holocene epoch.\",\n      \"volume\": \"431\",\n      \"identifier\": []\n    }\n  ],\n  \"studyName\": \"Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions\"\n}"
        },
        {
            "type": "csv",
            "filenameFull": "MD98-2170.Stott.2004/data/MD98-2170.Stott.2004.paleo1measurement1.csv",
            "filenameShort": "MD98-2170.Stott.2004.paleo1measurement1.csv",
            "data": {
                "data": [
                    [
                        "5.0",
                        "0.0",
                        "-2.828",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "7.5",
                        "96.2",
                        "-2.793",
                        "0.169",
                        "5.366",
                        "29.42"
                    ],
                    [
                        "10.0",
                        "192.3",
                        "-2.785",
                        "0.154",
                        "5.315",
                        "29.31"
                    ],
                    [
                        "12.5",
                        "288.5",
                        "-2.565",
                        "0.415",
                        "5.412",
                        "29.51"
                    ],
                    [
                        "15.0",
                        "384.6",
                        "-2.925",
                        "0.005",
                        "5.296",
                        "29.27"
                    ],
                    [
                        "17.5",
                        "480.8",
                        "-2.796",
                        "0.139",
                        "5.304",
                        "29.29"
                    ],
                    [
                        "20.0",
                        "576.9",
                        "-2.931",
                        "-0.005",
                        "5.284",
                        "29.25"
                    ],
                    [
                        "22.5",
                        "673.1",
                        "-2.955",
                        "0.021",
                        "5.401",
                        "29.49"
                    ],
                    [
                        "25.0",
                        "769.2",
                        "-2.879",
                        "0.039",
                        "5.266",
                        "29.21"
                    ],
                    [
                        "30.0",
                        "961.5",
                        "-2.983",
                        "0.018",
                        "5.458",
                        "29.61"
                    ],
                    [
                        "40.0",
                        "1346.2",
                        "-2.899",
                        "0.058",
                        "5.356",
                        "29.4"
                    ],
                    [
                        "50.0",
                        "1730.8",
                        "-2.756",
                        "0.254",
                        "5.477",
                        "29.65"
                    ],
                    [
                        "65.0",
                        "2307.7",
                        "-2.918",
                        "-0.004",
                        "5.258",
                        "29.19"
                    ],
                    [
                        "80.0",
                        "2884.6",
                        "-2.799",
                        "0.177",
                        "5.402",
                        "29.49"
                    ],
                    [
                        "90.0",
                        "3269.2",
                        "-2.808",
                        "0.174",
                        "5.413",
                        "29.52"
                    ],
                    [
                        "100.0",
                        "3653.8",
                        "-2.763",
                        "0.072",
                        "5.079",
                        "28.81"
                    ],
                    [
                        "110.0",
                        "4038.5",
                        "-2.81",
                        "0.108",
                        "5.264",
                        "29.21"
                    ],
                    [
                        "120.0",
                        "4423.1",
                        "-2.936",
                        "0.061",
                        "5.448",
                        "29.59"
                    ],
                    [
                        "130.0",
                        "4807.7",
                        "-2.583",
                        "0.212",
                        "4.996",
                        "28.62"
                    ],
                    [
                        "154.0",
                        "5730.8",
                        "-2.865",
                        "0.138",
                        "5.463",
                        "29.62"
                    ],
                    [
                        "162.0",
                        "6038.5",
                        "-2.757",
                        "0.23",
                        "5.426",
                        "29.54"
                    ],
                    [
                        "168.0",
                        "6269.2",
                        "-2.911",
                        "0.122",
                        "5.534",
                        "29.76"
                    ],
                    [
                        "174.0",
                        "6500.0",
                        "-2.78",
                        "0.25",
                        "5.53",
                        "29.75"
                    ],
                    [
                        "180.0",
                        "6730.8",
                        "NaN",
                        "NaN",
                        "5.494",
                        "29.68"
                    ],
                    [
                        "183.0",
                        "6846.2",
                        "-2.877",
                        "0.122",
                        "5.456",
                        "29.6"
                    ],
                    [
                        "200.0",
                        "7203.0",
                        "-2.833",
                        "0.093",
                        "5.288",
                        "29.25"
                    ],
                    [
                        "221.0",
                        "7897.8",
                        "-2.518",
                        "0.481",
                        "5.452",
                        "29.6"
                    ],
                    [
                        "240.0",
                        "8558.3",
                        "-2.667",
                        "0.416",
                        "5.655",
                        "30.0"
                    ],
                    [
                        "258.0",
                        "9212.1",
                        "-2.642",
                        "0.374",
                        "5.493",
                        "29.68"
                    ],
                    [
                        "283.0",
                        "10165.1",
                        "-2.563",
                        "0.553",
                        "5.736",
                        "30.16"
                    ],
                    [
                        "302.0",
                        "10670.6",
                        "-2.585",
                        "0.477",
                        "5.605",
                        "29.9"
                    ],
                    [
                        "308.0",
                        "10961.6",
                        "-2.588",
                        "0.503",
                        "5.676",
                        "30.04"
                    ],
                    [
                        "321.0",
                        "11469.9",
                        "-2.368",
                        "0.598",
                        "5.376",
                        "29.44"
                    ],
                    [
                        "327.0",
                        "11728.6",
                        "-2.552",
                        "0.383",
                        "5.305",
                        "29.29"
                    ],
                    [
                        "333.0",
                        "12034.3",
                        "-2.355",
                        "0.6",
                        "5.353",
                        "29.39"
                    ],
                    [
                        "340.0",
                        "12299.5",
                        "-2.107",
                        "0.759",
                        "5.151",
                        "28.96"
                    ],
                    [
                        "346.0",
                        "12567.8",
                        "-1.901",
                        "0.909",
                        "5.027",
                        "28.69"
                    ],
                    [
                        "358.0",
                        "13159.4",
                        "-1.868",
                        "0.996",
                        "5.146",
                        "28.95"
                    ],
                    [
                        "365.0",
                        "13437.3",
                        "-1.894",
                        "0.778",
                        "4.734",
                        "28.03"
                    ],
                    [
                        "377.0",
                        "14002.0",
                        "-1.756",
                        "0.849",
                        "4.601",
                        "27.71"
                    ],
                    [
                        "383.0",
                        "14288.9",
                        "-1.527",
                        "1.072",
                        "4.587",
                        "27.68"
                    ],
                    [
                        "389.0",
                        "14578.9",
                        "-1.024",
                        "1.502",
                        "4.447",
                        "27.33"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN",
                        "NaN"
                    ],
                    [
                        ""
                    ]
                ],
                "errors": [],
                "meta": {
                    "delimiter": ",",
                    "linebreak": "\r\n",
                    "cursor": 1825
                },
                "rows": 52,
                "cols": 6,
                "delimiter": ","
            },
            "pretty": "{\n  \"data\": [\n    [\n      \"5.0\",\n      \"0.0\",\n      \"-2.828\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"7.5\",\n      \"96.2\",\n      \"-2.793\",\n      \"0.169\",\n      \"5.366\",\n      \"29.42\"\n    ],\n    [\n      \"10.0\",\n      \"192.3\",\n      \"-2.785\",\n      \"0.154\",\n      \"5.315\",\n      \"29.31\"\n    ],\n    [\n      \"12.5\",\n      \"288.5\",\n      \"-2.565\",\n      \"0.415\",\n      \"5.412\",\n      \"29.51\"\n    ],\n    [\n      \"15.0\",\n      \"384.6\",\n      \"-2.925\",\n      \"0.005\",\n      \"5.296\",\n      \"29.27\"\n    ],\n    [\n      \"17.5\",\n      \"480.8\",\n      \"-2.796\",\n      \"0.139\",\n      \"5.304\",\n      \"29.29\"\n    ],\n    [\n      \"20.0\",\n      \"576.9\",\n      \"-2.931\",\n      \"-0.005\",\n      \"5.284\",\n      \"29.25\"\n    ],\n    [\n      \"22.5\",\n      \"673.1\",\n      \"-2.955\",\n      \"0.021\",\n      \"5.401\",\n      \"29.49\"\n    ],\n    [\n      \"25.0\",\n      \"769.2\",\n      \"-2.879\",\n      \"0.039\",\n      \"5.266\",\n      \"29.21\"\n    ],\n    [\n      \"30.0\",\n      \"961.5\",\n      \"-2.983\",\n      \"0.018\",\n      \"5.458\",\n      \"29.61\"\n    ],\n    [\n      \"40.0\",\n      \"1346.2\",\n      \"-2.899\",\n      \"0.058\",\n      \"5.356\",\n      \"29.4\"\n    ],\n    [\n      \"50.0\",\n      \"1730.8\",\n      \"-2.756\",\n      \"0.254\",\n      \"5.477\",\n      \"29.65\"\n    ],\n    [\n      \"65.0\",\n      \"2307.7\",\n      \"-2.918\",\n      \"-0.004\",\n      \"5.258\",\n      \"29.19\"\n    ],\n    [\n      \"80.0\",\n      \"2884.6\",\n      \"-2.799\",\n      \"0.177\",\n      \"5.402\",\n      \"29.49\"\n    ],\n    [\n      \"90.0\",\n      \"3269.2\",\n      \"-2.808\",\n      \"0.174\",\n      \"5.413\",\n      \"29.52\"\n    ],\n    [\n      \"100.0\",\n      \"3653.8\",\n      \"-2.763\",\n      \"0.072\",\n      \"5.079\",\n      \"28.81\"\n    ],\n    [\n      \"110.0\",\n      \"4038.5\",\n      \"-2.81\",\n      \"0.108\",\n      \"5.264\",\n      \"29.21\"\n    ],\n    [\n      \"120.0\",\n      \"4423.1\",\n      \"-2.936\",\n      \"0.061\",\n      \"5.448\",\n      \"29.59\"\n    ],\n    [\n      \"130.0\",\n      \"4807.7\",\n      \"-2.583\",\n      \"0.212\",\n      \"4.996\",\n      \"28.62\"\n    ],\n    [\n      \"154.0\",\n      \"5730.8\",\n      \"-2.865\",\n      \"0.138\",\n      \"5.463\",\n      \"29.62\"\n    ],\n    [\n      \"162.0\",\n      \"6038.5\",\n      \"-2.757\",\n      \"0.23\",\n      \"5.426\",\n      \"29.54\"\n    ],\n    [\n      \"168.0\",\n      \"6269.2\",\n      \"-2.911\",\n      \"0.122\",\n      \"5.534\",\n      \"29.76\"\n    ],\n    [\n      \"174.0\",\n      \"6500.0\",\n      \"-2.78\",\n      \"0.25\",\n      \"5.53\",\n      \"29.75\"\n    ],\n    [\n      \"180.0\",\n      \"6730.8\",\n      \"NaN\",\n      \"NaN\",\n      \"5.494\",\n      \"29.68\"\n    ],\n    [\n      \"183.0\",\n      \"6846.2\",\n      \"-2.877\",\n      \"0.122\",\n      \"5.456\",\n      \"29.6\"\n    ],\n    [\n      \"200.0\",\n      \"7203.0\",\n      \"-2.833\",\n      \"0.093\",\n      \"5.288\",\n      \"29.25\"\n    ],\n    [\n      \"221.0\",\n      \"7897.8\",\n      \"-2.518\",\n      \"0.481\",\n      \"5.452\",\n      \"29.6\"\n    ],\n    [\n      \"240.0\",\n      \"8558.3\",\n      \"-2.667\",\n      \"0.416\",\n      \"5.655\",\n      \"30.0\"\n    ],\n    [\n      \"258.0\",\n      \"9212.1\",\n      \"-2.642\",\n      \"0.374\",\n      \"5.493\",\n      \"29.68\"\n    ],\n    [\n      \"283.0\",\n      \"10165.1\",\n      \"-2.563\",\n      \"0.553\",\n      \"5.736\",\n      \"30.16\"\n    ],\n    [\n      \"302.0\",\n      \"10670.6\",\n      \"-2.585\",\n      \"0.477\",\n      \"5.605\",\n      \"29.9\"\n    ],\n    [\n      \"308.0\",\n      \"10961.6\",\n      \"-2.588\",\n      \"0.503\",\n      \"5.676\",\n      \"30.04\"\n    ],\n    [\n      \"321.0\",\n      \"11469.9\",\n      \"-2.368\",\n      \"0.598\",\n      \"5.376\",\n      \"29.44\"\n    ],\n    [\n      \"327.0\",\n      \"11728.6\",\n      \"-2.552\",\n      \"0.383\",\n      \"5.305\",\n      \"29.29\"\n    ],\n    [\n      \"333.0\",\n      \"12034.3\",\n      \"-2.355\",\n      \"0.6\",\n      \"5.353\",\n      \"29.39\"\n    ],\n    [\n      \"340.0\",\n      \"12299.5\",\n      \"-2.107\",\n      \"0.759\",\n      \"5.151\",\n      \"28.96\"\n    ],\n    [\n      \"346.0\",\n      \"12567.8\",\n      \"-1.901\",\n      \"0.909\",\n      \"5.027\",\n      \"28.69\"\n    ],\n    [\n      \"358.0\",\n      \"13159.4\",\n      \"-1.868\",\n      \"0.996\",\n      \"5.146\",\n      \"28.95\"\n    ],\n    [\n      \"365.0\",\n      \"13437.3\",\n      \"-1.894\",\n      \"0.778\",\n      \"4.734\",\n      \"28.03\"\n    ],\n    [\n      \"377.0\",\n      \"14002.0\",\n      \"-1.756\",\n      \"0.849\",\n      \"4.601\",\n      \"27.71\"\n    ],\n    [\n      \"383.0\",\n      \"14288.9\",\n      \"-1.527\",\n      \"1.072\",\n      \"4.587\",\n      \"27.68\"\n    ],\n    [\n      \"389.0\",\n      \"14578.9\",\n      \"-1.024\",\n      \"1.502\",\n      \"4.447\",\n      \"27.33\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\",\n      \"NaN\"\n    ],\n    [\n      \"\"\n    ]\n  ],\n  \"errors\": [],\n  \"meta\": {\n    \"delimiter\": \",\",\n    \"linebreak\": \"\\r\\n\",\n    \"aborted\": false,\n    \"truncated\": false,\n    \"cursor\": 1825\n  },\n  \"rows\": 52,\n  \"cols\": 6,\n  \"delimiter\": \",\"\n}"
        }
    ],
    []
]

sample_data_response = [
    {
        "dat": {
            "studyName": "Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions",
            "dataSetName": "MD98-2170.Stott.2004",
            "chronData": [],
            "paleoData": []
        },
        "feedback": {
            "missingTsidCt": 11,
            "wrnCt": 0,
            "errCt": 1,
            "tsidMsgs": [
                "Missing data: paleoData0paleoMeasurementTable0.column0.TSid",
            ],
            "posMsgs": [
                "Valid Bagit File"
            ],
            "errMsgs": [
                "Error message here"
            ],
            "wrnMsgs": []
        },
        "status": "PASS",
        "filename": ""

    },
    {
        "dat": {
            "studyName": "Western Tropical Pacific Holocene Sea Surface Temperature and Salinity Reconstructions",
            "dataSetName": "MD98-2170.Stott.2004",
            "chronData": [],
            "paleoData": []
        },
        "feedback": {
            "missingTsidCt": 11,
            "wrnCt": 0,
            "errCt": 1,
            "tsidMsgs": [
                "Missing data: paleoData0paleoMeasurementTable0.column0.TSid",
            ],
            "posMsgs": [
                "Valid Bagit File"
            ],
            "errMsgs": [
                "Error message here"
            ],
            "wrnMsgs": []
        },
        "status": "PASS",
        "filename": ""

    }

]



import os
import shutil

from ..helpers.directory import create_tmp_dir
from ..helpers.zips import zipper
from .lpd_noaa import LPD_NOAA
# from .noaa_lpd import NOAA_LPD
from ..helpers.loggers import create_logger


logger_noaa = create_logger("noaa")


def noaa_prompt():
    """
    Convert between NOAA and LiPD file formats.
    :return:
    """
    logger_noaa.info("enter noaa")
    # Run lpd_noaa or noaa_lpd ?
    print("Which conversion?\n1. LPD to NOAA\n2. NOAA to LPD (under construction!)\n")
    mode = input("Option: ")
    logger_noaa.info("chose option: {}".format(mode))

    return mode


def noaa_to_lpd(files):
    """
    Convert NOAA format to LiPD format
    :param dict file: File metadata
    :param str dir_tmp: Path to tmp directory
    :return None:
    """
    logger_noaa.info("enter process_noaa")
    # only continue if the user selected a mode correctly
    logger_noaa.info("Found {} NOAA txt file(s)".format(str(len(files[".txt"]))))
    print("Found {} NOAA txt file(s)".format(str(len(files[".txt"]))))
    # Process each available file of the specified .lpd or .txt type
    for file in files[".txt"]:
        # try to filter out example files and stuff without real data
        if "template" not in file["filename_ext"] and "example" not in file["filename_ext"]:
            os.chdir(file["dir"])
            print('processing: {}'.format(file["filename_ext"]))
            logger_noaa.info("processing: {}".format(file["filename_ext"]))

            # Unzip file and get tmp directory path
            dir_tmp = create_tmp_dir()

            # NOAA_LPD(dir_root, dir_tmp, name).main()
            os.chdir(file["dir"])

            zipper(path_name_ext=file["filename_ext"], root_dir=dir_tmp, name=file["filename_no_ext"])

            # Delete tmp folder and all contents
            os.chdir(file["dir"])
            try:
                shutil.rmtree(dir_tmp)
            except FileNotFoundError:
                # directory is already gone. keep going.
                pass

    logger_noaa.info("exit noaa_to_lpd")
    return


def lpd_to_noaa(obj):
    """
    Convert a LiPD format to NOAA format
    :param obj obj: LiPD object
    :return obj: LiPD object (modified)
    """
    logger_noaa.info("enter process_lpd")

    try:
        os.chdir(obj.dir_root)
        # Get the json data from the lipd object
        d = obj.get_metadata()
        # Create the conversion object, and start the conversion process
        _convert_obj = LPD_NOAA(obj.dir_root, obj.name, d, obj.data_csv)
        _convert_obj.main()
        # add the URL to the json dict and master dict in the object.
        d["WDCPaleoUrl"] = _convert_obj.get_wdc_paleo_url()
        m = obj.get_master()
        m["WDCPaleoUrl"] = _convert_obj.get_wdc_paleo_url()
        obj.put_master(m)
        obj.put_metadata(d)

    except FileNotFoundError:
        logger_noaa.debug("process_lpd: FileNotFound: tmp directory not found")
        print("Error: Unable to process {}".format(obj.name))
    logger_noaa.info("exit process_lpd")
    return obj



import numpy as np

from ..helpers.loggers import create_logger

logger_inferred_data = create_logger("inferred_data")


def _get_age(columns):
    """
    Sift through table column data and find the "age" or "year" column. Return its "values" data.
    :param dict columns: Column data
    :return list age: Age values
    """
    age = []
    try:
        # Check for age first (exact match)
        if "age" in columns:
            # Save the values
            age = columns["age"]["values"]
        # Check for year second (exact match)
        elif "year" in columns:
            # Save the values
            age = columns["year"]["values"]
        # No year or age found, start searching more intently.
        else:
            # Loop through column variableNames
            for k, v in columns.items():
                # Check for age in the variableName (loose match)
                if "age" in k:
                    # Save the values
                    age = v["values"]
                # Check for year in variableName (loose match)
                elif "year" in k:
                    # Save the values
                    age = v["values"]
    # If we expected a dictionary, and didn't get one
    except AttributeError as e:
        logger_inferred_data.warn("get_age: AttributeError: {}".format(e))
    # If we were looking for values, and didn't get one
    except KeyError as e:
        logger_inferred_data.warn("get_age: KeyError: {}".format(e))
    # Fail-safe for other problems
    except Exception as e:
        logger_inferred_data.warn("get_age: Exception: {}".format(e))

    return age


def _get_resolution(age, values):
    """
    Calculates the resolution (res)
    Thanks Deborah!
    """
    res = []
    try:
        # Get the nan index from the values and remove from age
        age2 = age[np.where(~np.isnan(values))[0]]
        res = np.diff(age2)
    except IndexError as e:
        print("get_resolution: IndexError: {}".format(e))
    except Exception as e:
        logger_inferred_data.warn("get_resolution: Exception: {}".format(e))
    return res


def _get_inferred_data_column(column, age):
    """
    Column level: Start creating the new fields and calculating the new data.
    :param dict column: Column data
    :param list age: Age values
    :return dict column: Column data - modified
    """
    try:
        # Make sure that age and values are numpy arrays
        age = np.asarray(age)
        # Get the values for this column
        values = column["values"]
        # If we have values, keep going
        if values:
            # Turn the values into a numpy array
            values = np.asarray(values)
            # Use the values to create new entries and data
            column["hasMinValue"] = np.min(values).tolist()
            column["hasMaxValue"] = np.max(values).tolist()
            column["hasMedianValue"] = np.median(values).tolist()
            column["hasMeanValue"] = np.mean(values).tolist()
            # Get the resolution for this age and column values data
            res = _get_resolution(age, values)
            # If we have successful resolution data, keep going
            if len(res) != 0:
                # Use the resolution values to create new entries and data
                column["hasResolution"] = {
                    "hasMinValue": np.min(res).tolist(),
                    "hasMaxValue": np.max(res).tolist(),
                    "hasMeanValue": np.mean(res).tolist(),
                    "hasMedianValue": np.median(res).tolist(),
                    "values": res.tolist()
                }
    except KeyError as e:
        logger_inferred_data.debug("get_inferred_data_column: KeyError: {}".format(e))
    except Exception as e:
        logger_inferred_data.debug("get_inferred_data_column: Exception: {}".format(e))

    return column


def get_inferred_data_table(table):
    """
    Table level: Dive down, calculate data, then return the new table with the inferred data.
    :param dict table: Table data
    :return dict table: Table with new data
    """

    # Get the age values data first, since it's needed to calculate the other column data.
    age = _get_age(table["columns"])

    # If age values were not found, then don't continue. There's no point!
    if age:
        try:
            # Loop for all the columns in the table
            for var, col in table["columns"].items():
                # Special cases
                # We do not calculate data for each of the keys below, and we cannot calculate any "string" data
                if var not in ["depth", "age", "year"] and not all(isinstance(i, str) for i in col["values"]):
                    # Send this particular column down to be calculated. Set the returned data in-place in the table.
                    table["columns"][var] = _get_inferred_data_column(col, age)
                # If we want to know which columns aren't being calculated...they end up here.
                # else:
                #     logger_inferred_data.info("get_inferred_data_table: "
                #                               "Not calculating inferred data for variableName: {}".format(var))
        except AttributeError as e:
            logger_inferred_data.warn("get_inferred_data_table: AttributeError: {}".format(e))
        except Exception as e:
            logger_inferred_data.warn("get_inferred_data_table: Exception: {}".format(e))
    else:
        try:
            logger_inferred_data.info("Unable to calculate inferred data for: {}".format(table["filename"]))
        except KeyError:
            logger_inferred_data.warn("get_inferred_data_table: Unable to calculate inferred data for: unknown table")

    return table
